from typing import Union

from beancount.core.data import Transaction
from beancount.parser import parser


class Dispatcher:
    """
    交易语句处理器
    """

    def quick_check(self, input_str: str) -> bool:
        """
        快速检查输入是否符合
        :param input_str:
        :return:
        """
        return True

    def process(self, input_str: str) -> Union[Transaction, str]:
        """
        解析输入为交易。若输入不合规，则抛出 ValueError
        :param input_str:
        :return:
        """
        tx_str = self._process_raw(input_str)
        try:
            tx = parser.parse_one(tx_str)
            if isinstance(tx, Transaction):
                return tx
            else:
                return tx_str
        except AssertionError:
            return tx_str

    def _process_raw(self, input_str: str) -> str:
        """
        解析输入为交易字符串
        :param input_str:
        :return:
        """
        return '''
               2010-01-01 * "Payee" "Desc"
                 Assets:Unknown
                 Expenses:Unknown    + 1 CNY
               '''

    def get_name(self) -> str:
        """
        获得处理器名称。用于在 /help 中显示
        :return:
        """
        return '未知'

    def get_usage(self) -> str:
        """
        获得帮助信息
        :return:
        """
        return '暂无帮助信息。'



import copy
import datetime
import os
import time
import uuid
from typing import List, Tuple, Union

from beancount.core.data import Transaction
from beancount.parser import printer, parser

from beancount_bot.config import get_global, GLOBAL_MANAGER, get_config
from beancount_bot.dispatcher import Dispatcher
from beancount_bot.util import load_class

META_UUID = 'tgbot_uuid'
META_TIME = 'tgbot_time'

Uuid = str


class NotMatchException(Exception):
    pass


class TransactionManager:
    """
    交易信息管理
    """

    def __init__(self, dispatchers: List[Dispatcher], bean_file: str):
        self.dispatchers = dispatchers
        self.__bean_file = bean_file

    def create(self, tx: Union[Transaction, str]) -> Tuple[Uuid, Union[Transaction, str]]:
        """
        创建交易
        :param tx:
        :return:
        """
        tx_uuid = Uuid(uuid.uuid4())
        if isinstance(tx, str):
            # 保存至账本
            with open(self.bean_file, 'a+', encoding='utf-8') as f:
                f.write(f"; TGBOT_START {tx_uuid}\n{tx}\n; TGBOT_END {tx_uuid}\n")
            return tx_uuid, tx
        elif isinstance(tx, Transaction):
            # 添加控制元数据
            tx = copy.deepcopy(tx)
            tx.meta[META_UUID] = tx_uuid
            tx.meta[META_TIME] = str(datetime.datetime.now())
            # 保存至账本
            with open(self.bean_file, 'a+', encoding='utf-8') as f:
                printer.print_entry(tx, file=f)
            return tx_uuid, tx
        else:
            raise ValueError()

    def remove(self, tx_uuid: Uuid) -> Union[Transaction, str]:
        """
        删除交易
        :param tx_uuid:
        :return:
        """
        entries, errors, _ = parser.parse_file(self.bean_file)
        if len(errors) > 0:
            desc = '\n'.join(map(lambda err: f'行 {err.source["lineno"]}：{err.message}', errors))
            raise ValueError("账本文件内容错误！\n" + desc)
        # 筛选交易
        to_delete = next(
            filter(lambda tx: tx.meta[META_UUID] == tx_uuid if META_UUID in tx.meta else False, entries),
            None
        )
        if to_delete is None:
            # 可能是非交易语句
            return self._remove_comment_wrapped(tx_uuid)
        # 统计删除行。避免删除其他语句。
        min_line = to_delete.meta['lineno']
        max_line = min_line
        for posting in to_delete.postings:
            max_line = max(max_line, posting.meta['lineno'])
        # 删除
        with open(self.bean_file, 'r', encoding='utf-8') as f:
            lines = f.readlines()
        with open(self.bean_file, 'w', encoding='utf-8') as f:
            f.write(''.join(lines[:min_line - 1] + lines[max_line:]))
        return to_delete

    def _remove_comment_wrapped(self, tx_uuid: Uuid) -> str:
        """
        使用注释包裹
        :param tx_uuid:
        :return:
        """
        with open(self.bean_file, 'r', encoding='utf-8') as f:
            lines = f.readlines()
        # 筛选列
        min_line = -1
        for i in range(len(lines)):
            if f'TGBOT_START {tx_uuid}' in lines[i]:
                min_line = i
                break
        max_line = -1
        for i in range(len(lines)):
            if f'TGBOT_END {tx_uuid}' in lines[i]:
                max_line = i
                break
        if min_line == -1 or max_line == -1:
            raise ValueError("交易不存在！")
        # 删除
        with open(self.bean_file, 'w', encoding='utf-8') as f:
            f.write(''.join(lines[:min_line] + lines[max_line + 1:]))
        return ''.join(lines[min_line + 1:max_line])[:-1]

    def create_from_str(self, tx_str) -> Union[Tuple[Uuid, Transaction], Tuple[None, str]]:
        """
        从交易语法创建交易
        :param tx_str:
        :return:
        """
        tx = self._parse_transaction(tx_str)
        tx_uuid, _ = self.create(tx)
        return tx_uuid, tx

    def _parse_transaction(self, tx_str) -> Transaction:
        for dispatcher in self.dispatchers:
            if not dispatcher.quick_check(tx_str):
                continue
            # 尝试解析
            try:
                tx = dispatcher.process(tx_str)
                return tx
            except NotMatchException:
                # 不能通过该解析器解析
                continue
        else:
            # 没有匹配
            raise ValueError("无法识别此交易语法")

    @property
    def bean_file(self) -> str:
        params = {
            'year': time.strftime("%Y", time.localtime()),
            'month': time.strftime("%m", time.localtime()),
            'date': time.strftime("%d", time.localtime()),
        }
        bean_file = self.__bean_file
        for k, v in params.items():
            bean_file = bean_file.replace(f'{{{k}}}', v)
        # 创建父文件夹
        path = os.path.dirname(os.path.realpath(bean_file))
        os.makedirs(path, exist_ok=True)
        return bean_file


def stringfy(tx: Union[Transaction, str]) -> str:
    """
    交易转为字符串
    :param tx:
    :return:
    """
    if isinstance(tx, str):
        return tx
    return printer.format_entry(tx)


def get_manager() -> TransactionManager:
    """
    从配置创建管理对象
    :return:
    """

    def create_manager():
        # 创建分发器
        dispatchers = []
        for conf in get_config('transaction.message_dispatcher', []):
            clazz = load_class(conf['class'])
            dispatchers.append(clazz(**conf['args']))
        # 获得 Bean 文件位置
        bean_file: str = get_config('transaction.beancount_file')
        # 创建对象
        return TransactionManager(dispatchers, bean_file)

    return get_global(GLOBAL_MANAGER, create_manager)



import telebot
from telebot import apihelper
from telebot.types import InlineKeyboardMarkup, InlineKeyboardButton, MessageEntity, Message, CallbackQuery

from beancount_bot import transaction
from beancount_bot.config import get_config, load_config
from beancount_bot.session import get_session, SESS_AUTH, get_session_for, set_session
from beancount_bot.task import load_task, get_task
from beancount_bot.transaction import get_manager
from beancount_bot.util import logger

apihelper.ENABLE_MIDDLEWARE = True

bot = telebot.TeleBot(token=None, parse_mode=None)


@bot.middleware_handler(update_types=['message'])
def session_middleware(bot_instance, message):
    """
    会话中间件
    :param bot_instance:
    :param message:
    :return:
    """
    bot_instance.session = get_session_for(message.from_user.id)


#######
# 鉴权 #
#######

def check_auth() -> bool:
    """
    检查是否登录
    :return:
    """
    return SESS_AUTH in bot.session and bot.session[SESS_AUTH]


@bot.message_handler(commands=['start'])
def start_handler(message: Message):
    """
    首次聊天时鉴权
    :param message:
    :return:
    """
    auth = get_session(message.from_user.id, SESS_AUTH, False)
    if auth:
        bot.reply_to(message, "已经鉴权过了！")
        return
    # 要求鉴权
    bot.reply_to(message, "欢迎使用记账机器人！请输入鉴权令牌：")


def auth_token_handler(message: Message):
    """
    登陆令牌回调
    :param message:
    :return:
    """
    if check_auth():
        return
    # 未鉴权则视为鉴权令牌
    auth_token = get_config('bot.auth_token')
    if auth_token == message.text:
        set_session(message.from_user.id, SESS_AUTH, True)
        bot.reply_to(message, "鉴权成功！")
    else:
        bot.reply_to(message, "鉴权令牌错误！")


#######
# 指令 #
#######


@bot.message_handler(commands=['reload'])
def reload_handler(message):
    """
    重载配置指令
    :param message:
    :return:
    """
    if not check_auth():
        bot.reply_to(message, "请先进行鉴权！")
        return
    load_config()
    load_task()
    bot.reply_to(message, "成功重载配置！")


@bot.message_handler(commands=['help'])
def help_handler(message):
    """
    帮助指令
    :param message:
    :return:
    """
    # 创建消息按键
    markup = InlineKeyboardMarkup()
    dispatchers = get_manager().dispatchers
    for ind, d in zip(range(len(dispatchers)), dispatchers):
        markup.add(InlineKeyboardButton(f'帮助：{d.get_name()}', callback_data=f'help:{ind}'))
    # 帮助信息
    command_usage = [
        '/start - 鉴权',
        '/help - 使用帮助',
        '/reload - 重新加载配置文件',
        '/task - 查看、运行任务',
    ]
    bot.reply_to(message, "记账 Bot\n\n可用指令列表：\n"
                 + '\n'.join(command_usage)
                 + "\n\n交易语句帮助请选择对应模块。", reply_markup=markup)


@bot.callback_query_handler(func=lambda call: call.data[:4] == 'help')
def callback_help(call: CallbackQuery):
    try:
        d_id = int(call.data[5:])
        dispatchers = get_manager().dispatchers
        d = dispatchers[d_id]
        bot.reply_to(call.message, f'帮助：{d.get_name()}\n\n' + d.get_usage())
    except Exception as e:
        logger.error(f'{call.id}：发生未知错误！', e)
        bot.answer_callback_query(call.id, '发生未知错误！')


@bot.message_handler(commands=['task'])
def task_handler(message):
    """
    任务指令
    :param message:
    :return:
    """
    if not check_auth():
        bot.reply_to(message, "请先进行鉴权！")
        return

    cmd = message.text
    tasks = get_task()
    if cmd == '/task':
        # 显示所有任务
        bot.reply_to(message, '当前注册任务：' + ', '.join(tasks.keys()))
    else:
        # 运行任务
        dest = cmd[6:]
        task = tasks[dest]
        task.trigger(bot)


#######
# 交易 #
#######


@bot.message_handler(func=lambda m: True)
def transaction_query_handler(message: Message):
    """
    交易语句处理
    :param message:
    :return:
    """
    if not check_auth():
        auth_token_handler(message)
        return
    # 已鉴权则处理交易
    manager = get_manager()
    try:
        tx_uuid, tx = manager.create_from_str(message.text)
        # 创建消息按键
        markup = InlineKeyboardMarkup()
        markup.add(InlineKeyboardButton('撤回交易', callback_data=f'withdraw:{tx_uuid}'))
        # 回复
        bot.reply_to(message, transaction.stringfy(tx), reply_markup=markup)
    except ValueError as e:
        logger.info(f'{message.from_user.id}：无法添加交易', e)
        bot.reply_to(message, e.args[0])
    except Exception as e:
        logger.error(f'{message.from_user.id}：发生未知错误！添加交易失败。', e)
        bot.reply_to(message, '发生未知错误！添加交易失败。')


@bot.callback_query_handler(func=lambda call: call.data[:8] == 'withdraw')
def callback_withdraw(call: CallbackQuery):
    """
    交易撤回回调
    :param call:
    :return:
    """
    auth = get_session(call.from_user.id, SESS_AUTH, False)
    if not auth:
        bot.answer_callback_query(call.id, '请先进行鉴权！')
        return
    tx_uuid = call.data[9:]
    manager = get_manager()
    try:
        manager.remove(tx_uuid)
        # 修改原消息回复
        message = '交易已撤回'
        code_format = MessageEntity('code', 0, len(message))
        bot.edit_message_text(message,
                              chat_id=call.message.chat.id,
                              message_id=call.message.message_id,
                              entities=[code_format])
    except ValueError as e:
        logger.info(f'{call.id}：无法创建交易', e)
        bot.answer_callback_query(call.id, e.args[0])
    except Exception as e:
        logger.error(f'{call.id}：发生未知错误！撤回交易失败。', e)
        bot.answer_callback_query(call.id, '发生未知错误！撤回交易失败。')


def serving():
    """
    启动 Bot
    :return:
    """

    # 设置 Token
    token = get_config('bot.token')
    bot.token = token
    # 设置代理
    proxy = get_config('bot.proxy')
    if proxy is not None:
        apihelper.proxy = {'https': proxy}
    # 启动
    bot.infinity_polling()



import unittest

from beancount_bot import transaction
from beancount_bot.dispatcher import Dispatcher


class TestDispatcher(unittest.TestCase):

    def test_process(self):
        dispatcher = Dispatcher()
        tx = dispatcher.process('')
        ret = transaction.stringfy(tx)
        self.assertEqual(ret,
                         '2010-01-01 * "Payee" "Desc"\n'
                         '  Assets:Unknown\n'
                         '  Expenses:Unknown  1 CNY\n')



import sys

import telebot

logger = telebot.logger


def load_class(classname: str) -> type:
    """
    通过类名加载类
    :param classname:
    :return:
    """
    class_path = classname.split('.')
    module, classname = '.'.join(class_path[:-1]), class_path[-1]
    __import__(module)
    return getattr(sys.modules[module], classname)



from beancount_bot import builtin
from beancount_bot import dispatcher
from beancount_bot import task
from beancount_bot import transaction
from beancount_bot import util
from beancount_bot.main import main

__AUTHOR__ = 'KAAAsS'
__LICENSE__ = "MIT"
__VERSION__ = "1.1.4"



import click

from beancount_bot import bot, config as conf
from beancount_bot.config import load_config, get_config
from beancount_bot.session import load_session
from beancount_bot.task import load_task, start_schedule_thread
from beancount_bot.transaction import get_manager
from beancount_bot.util import logger


@click.command()
@click.option('-c', '--config', default='beancount_bot.yml', help='配置文件路径')
def main(config):
    """
    适用于 Beancount 的 Telegram 机器人
    """
    logger.setLevel('INFO')
    # 加载配置
    logger.info("加载配置：%s", config)
    conf.config_file = config
    load_config()
    # 设置日志等级
    logger.setLevel(get_config('log.level', 'INFO'))
    # 加载会话
    logger.info("加载会话...")
    load_session()
    # 创建管理对象
    logger.info("创建管理对象...")
    get_manager()
    # 加载定时任务
    logger.info("加载定时任务...")
    load_task()
    start_schedule_thread()
    # 启动
    logger.info("启动 Bot...")
    bot.serving()


if __name__ == '__main__':
    main()



import threading
import time
from typing import Dict

import schedule
from telebot import TeleBot

from beancount_bot.config import get_config, get_global, GLOBAL_TASK
from beancount_bot.util import logger, load_class

_schedule_thread: threading.Thread = None


class ScheduleTask:
    """
    定时任务
    """

    def __init__(self):
        self.config = None

    def register(self, fire: callable):
        """
        注册定时任务
        :param fire: 待执行函数
        :return:
        """
        pass

    def trigger(self, bot: TeleBot):
        """
        触发任务
        :param bot: Bot 对象
        :return:
        """
        pass


def load_task() -> Dict[str, ScheduleTask]:
    """
    加载定时任务
    :return:
    """
    from beancount_bot.bot import bot
    ret = {}
    schedule.clear()
    for conf in get_config('schedule', []):
        name = conf['name']
        clazz = load_class(conf['class'])
        args = conf['args']

        logger.info('注册定时任务：%s', name)
        task: ScheduleTask = clazz(**args)
        task.register(lambda: task.trigger(bot))
        task.config = conf

        ret[name] = task
    return ret


def get_task() -> Dict[str, ScheduleTask]:
    """
    获得任务
    :return:
    """
    return get_global(GLOBAL_TASK, load_task)


def start_schedule_thread(interval=5) -> threading.Thread:
    """
    运行定时任务
    :param interval:
    :return:
    """

    class ScheduleThread(threading.Thread):
        @classmethod
        def run(cls):
            while True:
                schedule.run_pending()
                time.sleep(interval)

    global _schedule_thread
    _schedule_thread = ScheduleThread()
    _schedule_thread.setDaemon(True)
    _schedule_thread.start()
    return _schedule_thread



import yaml

global_object_map = {}

GLOBAL_CONFIG = 'config'
GLOBAL_MANAGER = 'manager'
GLOBAL_TASK = 'task'

config_file = ''


def set_global(key: str, obj):
    """
    设置全局对象
    :param key:
    :param obj:
    :return:
    """
    global global_object_map
    global_object_map[key] = obj


def get_global(key: str, default_producer: callable):
    """
    获取全局对象
    :param key:
    :param default_producer:
    :return:
    """
    global global_object_map
    if key not in global_object_map:
        set_global(key, default_producer())
    return global_object_map[key]


def load_config(path=None):
    """
    从文件加载配置，将清空全局对象
    :param path:
    :return:
    """
    if path is None:
        path = config_file
    global global_object_map
    with open(path, 'r', encoding='utf-8') as f:
        data = yaml.full_load(f)
    global_object_map = {}
    set_global(GLOBAL_CONFIG, data)


def get_config_obj():
    """
    获得配置对象
    :return:
    """

    def _exception():
        raise ValueError('配置未载入！')

    return get_global(GLOBAL_CONFIG, _exception)


def get_config(key_path: str, default_value=None):
    """
    获得配置
    :param key_path:
    :param default_value:
    :return:
    """
    obj = get_config_obj()
    for ind in key_path.split('.'):
        if ind not in obj:
            return default_value
        obj = obj[ind]
    return obj



import json
import os.path
from types import MappingProxyType
from typing import Dict, Iterable

from beancount_bot.config import get_config
from beancount_bot.util import logger

SESS_AUTH = 'auth'

_session_cache: Dict[str, dict] = {}


def load_session():
    """
    从文件恢复会话数据
    :return:
    """
    global _session_cache
    session_file = get_config('bot.session_file')
    if os.path.exists(session_file):
        with open(session_file, 'r', encoding='utf-8') as f:
            _session_cache = json.load(f)
        logger.debug("从文件恢复会话 %s", _session_cache)


def get_session_for(uid: int) -> MappingProxyType:
    """
    返回用户会话的不可变视图
    :param uid:
    :return:
    """
    uid = str(uid)
    if uid not in _session_cache:
        _session_cache[uid] = {}
    return MappingProxyType(_session_cache[uid])


def get_session(uid: int, key: str, default_value=None) -> object:
    """
    返回用户会话的某一值
    :param uid:
    :param key:
    :param default_value:
    :return:
    """
    uid = str(uid)
    if uid not in _session_cache:
        _session_cache[uid] = {}
    if key not in _session_cache[uid]:
        return default_value
    return _session_cache[uid][key]


def set_session(uid: int, key: str, value: object):
    """
    设置用户会话值
    :param uid:
    :param key:
    :param value:
    :return:
    """
    uid = str(uid)
    if uid not in _session_cache:
        _session_cache[uid] = {}
    _session_cache[uid][key] = value
    # 保存缓存
    session_file = get_config('bot.session_file')
    with open(session_file, 'w', encoding='utf-8') as f:
        json.dump(_session_cache, f)


def all_user(auth=True) -> Iterable[int]:
    """
    获得所有用户
    :param auth:
    :return:
    """
    if auth:
        return map(lambda t: int(t[0]), filter(lambda t: SESS_AUTH in t[1] and t[1][SESS_AUTH], _session_cache.items()))
    else:
        return map(int, _session_cache.keys())



import datetime
import itertools
from typing import List, Mapping

import yaml

from beancount_bot.dispatcher import Dispatcher
from beancount_bot.transaction import NotMatchException
from beancount_bot.util import logger

_CH_CLASS = [' ', '\"', '\\', '<']
_STATE_MAT = [
    # 空, ", \, <, 其他字符
    [0, 2, -1, 4, 1],  # 0: 空格
    [0, 2, -1, 4, 1],  # 1: 词
    [2, 0, 3, 2, 2],  # 2: 字符串
    [2, 2, 2, 2, 2],  # 3: 转义
    [0, 2, -1, -1, 1],  # 4: 符号
]


def split_command(cmd):
    """
    切分输入指令。按照空格分割，允许使用双引号字符串、反斜杠转义
    :param cmd:
    :return:
    """
    state = 0
    words: List[str] = []

    for i in range(len(cmd)):
        ch = cmd[i]
        # 字符类
        if ch in _CH_CLASS:
            ch_class = _CH_CLASS.index(ch)
        else:
            ch_class = 4
        # 状态转移
        state, old_state = _STATE_MAT[state][ch_class], state
        if state == -1:
            raise ValueError(f"位置 {i}：语法错误！不应出现符号 {ch}。")
        # 进入事件
        if state != old_state and old_state != 3:
            if state in [1, 2, 4]:
                words.append('')
            if state in [2, 3]:
                continue
        # 状态事件
        if state != 0:
            words[-1] += ch
    if state not in [0, 1, 4]:
        raise ValueError(f"位置 {len(cmd)}：语法错误！字符串、转义未结束。")
    return words


def _to_list(el):
    if isinstance(el, list):
        return el
    return [el]


Template = Mapping


def print_one_usage(template: Template) -> str:
    """
    打印一个模板的语法提示
    :param template:
    :return:
    """
    usage = ''
    # 指令
    command = template['command']
    if isinstance(command, list):
        usage += '(' + '|'.join(command) + ')'
    else:
        usage += command
    # 参数
    if 'args' in template:
        usage += ' ' + ' '.join(template['args'])
    # 可选参数
    if 'optional_args' in template:
        usage += ' ' + ' '.join(map(lambda s: f'[{s}]', template['optional_args']))
    return usage


class TemplateDispatcher(Dispatcher):
    """
    模板处理器。通过 Json 模板生成交易信息。
    """

    def get_name(self) -> str:
        return '模板'

    def get_usage(self) -> str:
        if len(self.templates) > 0:
            command_usage = '\n'.join([f'  - {print_one_usage(t)}' for t in self.templates])
        else:
            command_usage = '没有定义任何模板'

        default_account = self.config['default_account']

        if len(self.config['accounts']) > 0:
            account_alias = '\n'.join([f'  {k} - {v}' for k, v in self.config['accounts'].items()])
        else:
            account_alias = '没有定义账户'

        return '模板指令格式：指令名 必填参数 [可选参数] < 目标账户\n' \
               '  1. 指令名可以有多个，记为”(指令名1|指令名2|...)“；\n' \
               '  2. 目标账户可以省略。省略将使用默认账户\n\n' \
               f'当前定义的模板：\n{command_usage}\n\n' \
               f'默认账户：{default_account}\n支持的账户：\n{account_alias}'

    def __init__(self, template_config: str):
        with open(template_config, 'r', encoding='utf-8') as f:
            data = yaml.full_load(f)
        self.config = data['config']
        self.templates = data['templates']

    def quick_check(self, input_str: str) -> bool:
        words = split_command(input_str)
        prefixes = map(lambda t: _to_list(t['command']), self.templates)
        prefixes = itertools.chain(*prefixes)
        # 开头相同且有空格隔开
        return any(map(lambda prefix: words[0] == prefix, prefixes))

    def _process_raw(self, input_str: str) -> str:
        words = split_command(input_str)
        cmd, args = words[0], words[1:]
        # 选择模板
        template = next(
            filter(lambda t: cmd in _to_list(t['command']), self.templates),
            None
        )
        if template is None:
            raise NotMatchException()
        # 默认参数
        arg_map = {
            'account': self.config['default_account'],
            'date': datetime.date.today().isoformat(),
            'command': cmd,
        }
        # 解析目标账户（<语法）
        if '<' in args:
            split_at = args.index('<')
            args, account = args[:split_at], args[split_at + 1:]
            if len(account) != 1:
                raise ValueError("语法错误！不支持多目标账户。")
            arg_map['account'] = self.config['accounts'][account[0]]
        # 参数获取
        if 'args' in template:
            args_need = template['args']
            if len(args) < len(args_need):
                raise ValueError('参数过少！语法：' + print_one_usage(template))
            arg_map.update({k: v for k, v in zip(args_need, args)})
            args = args[len(args_need):]
        if 'optional_args' in template:
            optional_args = template['optional_args']
            if len(args) > len(optional_args):
                raise ValueError('参数过多！语法：' + print_one_usage(template))
            arg_map.update({k: v for k, v in zip(optional_args, args)})
            for empty_k in optional_args[len(args):]:
                arg_map[empty_k] = ''
            args = args[len(optional_args):]
        if len(args) != 0:
            raise ValueError('参数过多！语法：' + print_one_usage(template))
        # 计算待计算参数
        if 'computed' in template:
            for k, expr in template['computed'].items():
                arg_map[k] = eval(expr, None, arg_map)
        # 进行模板替换
        logger.debug('模板参数 %s', arg_map)
        ret = template['template']
        for k, v in arg_map.items():
            ret = ret.replace(f'{{{k}}}', str(v))
        return ret



import tempfile
import unittest
import uuid

from beancount_bot import transaction
from beancount_bot.dispatcher import Dispatcher
from beancount_bot.transaction import TransactionManager


class TestTransactionManager(unittest.TestCase):

    def setUp(self):
        with tempfile.NamedTemporaryFile('w+b', suffix='.bean', delete=False) as f:
            self.tmp_file = f.name

    def test_create(self):
        # Mock
        class MockDispatcher(Dispatcher):
            def _process_raw(self, input_str: str) -> str:
                return '''
                2010-01-01 * "Payee" "Desc"
                  Income:Unknown
                  Assets:Unknown  1 CNY
                '''

        # 测试
        manager = TransactionManager([MockDispatcher()], self.tmp_file)
        tx_uuid, tx = manager.create_from_str('')

        with open(self.tmp_file, 'r', encoding='utf-8') as f:
            data = f.read()
            self.assertIn(transaction.META_TIME, data)
            self.assertIn(transaction.META_UUID, data)

        tx_str = transaction.stringfy(tx)
        self.assertIn('2010-01-01 * "Payee" "Desc"\n', tx_str)
        self.assertIn('Income:Unknown\n', tx_str)
        self.assertIn('Assets:Unknown  1 CNY\n', tx_str)

    def test_create_raw(self):
        # Mock
        class MockDispatcher(Dispatcher):
            def _process_raw(self, input_str: str) -> str:
                return '; comment'

        # 测试
        manager = TransactionManager([MockDispatcher()], self.tmp_file)
        tx_uuid, tx = manager.create_from_str('')
        self.assertEqual(tx, '; comment')
        with open(self.tmp_file, 'r', encoding='utf-8') as f:
            self.assertIn(tx_uuid, f.read())

    def test_remove(self):
        # Mock
        poison_str = str(uuid.uuid4())

        class MockDispatcher(Dispatcher):
            def _process_raw(self, input_str: str) -> str:
                return f'''
                2010-01-01 * "Payee" "{poison_str}"
                  Income:Unknown
                  Assets:Unknown  1 CNY
                '''

        manager = TransactionManager([MockDispatcher()], self.tmp_file)
        with open(self.tmp_file, 'a+', encoding='utf-8') as f:
            pre = str(uuid.uuid4())
            f.write(f'; {pre}\n')
        tx_uuid, tx = manager.create_from_str('')
        with open(self.tmp_file, 'a+', encoding='utf-8') as f:
            post = str(uuid.uuid4())
            f.write(f'; {post}\n')
        # 删除
        delete_tx = manager.remove(tx_uuid)
        self.assertEqual(tx_uuid, delete_tx.meta[transaction.META_UUID])

        with open(self.tmp_file, 'r', encoding='utf-8') as f:
            data = f.read()
        self.assertNotIn(poison_str, data)
        self.assertIn(pre, data)
        self.assertIn(post, data)

    def test_remove_raw(self):
        # Mock
        class MockDispatcher(Dispatcher):
            def _process_raw(self, input_str: str) -> str:
                return '; comment1\n; comment2'

        manager = TransactionManager([MockDispatcher()], self.tmp_file)
        with open(self.tmp_file, 'a+', encoding='utf-8') as f:
            pre = str(uuid.uuid4())
            f.write(f'; {pre}\n')
        tx_uuid, tx = manager.create_from_str('')
        with open(self.tmp_file, 'a+', encoding='utf-8') as f:
            post = str(uuid.uuid4())
            f.write(f'; {post}\n')
        # 删除
        delete_tx = manager.remove(tx_uuid)
        self.assertEqual(tx, delete_tx)

        with open(self.tmp_file, 'r', encoding='utf-8') as f:
            data = f.read()
        self.assertNotIn('comment1', data)
        self.assertNotIn('comment2', data)
        self.assertNotIn(tx_uuid, data)
        self.assertIn(pre, data)
        self.assertIn(post, data)



from setuptools import setup

from beancount_bot import __VERSION__

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='beancount-bot',
    version=__VERSION__,
    packages=['beancount_bot', 'beancount_bot.builtin'],
    url='https://github.com/kaaass/beancount_bot',
    entry_points={
        "console_scripts": [
            'beancount_bot = beancount_bot:main'
        ]
    },
    install_requires=install_requires,
    python_requires='>=3.6.0',
    license='MIT',
    author='KAAAsS',
    author_email='admin@kaaass.net',
    description='A telegram bot designed for Beancount',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import os
from typing import List

import schedule
from telebot import TeleBot

from beancount_bot.session import all_user
from beancount_bot.task import ScheduleTask


class DailyCommandTask(ScheduleTask):
    """
    每日执行指令任务
    """

    def __init__(self, time: str, commands: List[str], message: str):
        """
        :param time: 每日更新时间
        :param commands: 执行指令
        :param message: 执行完成后发送信息
        """
        super().__init__()
        self.time = time
        self.commands = commands
        self.message = message

    def register(self, fire: callable):
        schedule.every().day.at(self.time).do(fire)

    def trigger(self, bot: TeleBot):
        # 执行指令
        for cmd in self.commands:
            os.system(cmd)
        # 发送信息
        for uid in all_user():
            bot.send_message(uid, self.message)
